#ifndef TARGET_DC
#include <stdio.h>
#include <stdlib.h>
#else
#include "dream.h"
#include "libc.h"
#include "dc.h"
#endif

#include "types.h"
#include "gleam!.h"
#include "fileio.h"
#include "mem.h"
#include "g6502.h"
#include "tables.h"


g6502_cpu gleam_cpu;

void g6502_reset()
{
	fprintf(stdout, "-> g6502_reset()\n");
	gleam_cpu.pc = nes_read_word(RESET_VECTOR);
	gleam_cpu.p = 0x20;
	gleam_cpu.sp = 0xff;
	gleam_cpu.cycle_count = 0;
	fprintf(stdout, ":: Reset Vector: [0x%04X]\n", gleam_cpu.pc);
	fprintf(stdout, ":: NMI Vector: [%04X]\n",
		nes_read_word(NMI_VECTOR));
	fprintf(stdout, ":: BRK Vector: [%04X]\n",
		nes_read_word(BREAK_VECTOR));
#ifdef DEBUG
	{
		u16 i, j = 0;
		for (i = 0; i < 256; i++)
			if (jtable[i] != g_unknown)
				j++;
		fprintf(stdout, ":: Opcode count: [%d]\n", j);
	}
#endif
	fprintf(stdout, "<- g6502_reset();\n");
}

int g6502_step()
{
	u8 opc = 0;

	opc = nes_read_byte(gleam_cpu.pc);

#ifdef DEBUG
	fprintf(stdout, ":: Executing: [0x%02X:0x%02X] [%s] ",
		gleam_cpu.pc, opc, onames[opc]);
#endif

	if (jtable[opc] != g_unknown)
		gleam_cpu.pc++;
	jtable[opc] ();


	if (jtable[opc] == g_unknown)
		return 0;
	else
		return 1;
}


void set_flags8(u8 val, u8 what)
{
	if (what & FL_NEGATIVE) {
		if (val & 0x80)
			SET_FLAG(gleam_cpu, FL_NEGATIVE, 1);
		else
			SET_FLAG(gleam_cpu, FL_NEGATIVE, 0);
	}
	if (what & FL_ZERO) {
		if (!val)
			SET_FLAG(gleam_cpu, FL_ZERO, 1);
		else
			SET_FLAG(gleam_cpu, FL_ZERO, 0);
	}
}

void stack_write_byte(u8 val)
{
	nes_write_byte(gleam_cpu.sp + 0x0100, val);
	gleam_cpu.sp--;
}

void stack_write_word(u16 val)
{
	stack_write_byte((val & 0xff00) >> 8);
	stack_write_byte(val & 0x00ff);
}

u8 stack_read_byte()
{
	gleam_cpu.sp++;
	return nes_read_byte(gleam_cpu.sp + 0x0100);
}

u16 stack_read_word()
{
	return (stack_read_byte() | (stack_read_byte() << 8));
}

u8 immediate()
{
	u8 val = nes_read_byte(gleam_cpu.pc);
	gleam_cpu.pc += 1;
	return val;
}

u16 immediate_word()
{
	u16 val = nes_read_word(gleam_cpu.pc);
	gleam_cpu.pc += 2;
	return val;
}

u8 absolute()
{
	u16 val = nes_read_word(gleam_cpu.pc);
	gleam_cpu.pc += 2;
	return nes_read_byte(val);
}

u16 relative()
{
	s16 cur = gleam_cpu.pc + 1;
	s8 val = nes_read_byte(gleam_cpu.pc);
	gleam_cpu.pc++;
	cur = ((signed) cur) + ((signed) val);
	return (unsigned) cur;
}

u8 zeropage()
{
	u8 val = nes_read_byte(gleam_cpu.pc);
	gleam_cpu.pc++;
	return nes_read_byte((val & 0xFF));
}

u16 post_indexed_indirect()
{
	u8 zp = nes_read_byte(gleam_cpu.pc);
	u16 addr = nes_read_word((u16) (zp & 0xFF));

	gleam_cpu.pc++;

	if (((addr & 0xff) + gleam_cpu.y) > 0x100)
		CYCLES(gleam_cpu, 1);

	addr += gleam_cpu.y;

	return addr;
}

void g_brk_00()
{
	gleam_cpu.pc++;
	stack_write_word(gleam_cpu.pc);
	SET_FLAG(gleam_cpu, FL_SINT, 1);
	stack_write_word(gleam_cpu.p);
	SET_FLAG(gleam_cpu, FL_INT, 1);
	SET_FLAG(gleam_cpu, FL_DEC, 0);
	gleam_cpu.pc = nes_read_word(BREAK_VECTOR);

#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	CYCLES(gleam_cpu, 7);
}

void g_ora_05()
{
	u8 imm = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", imm);
#endif
	gleam_cpu.a |= nes_read_byte(imm & 0xFF);
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 3);
}

void g_asl_06()
{
	u8 imm = immediate();
	u8 val = nes_read_byte(imm);
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", imm);
#endif

	if (val & 0x80)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	val <<= 1;
	nes_write_byte(imm, val);
	set_flags8(val, FL_NEGATIVE | FL_ZERO);
	CYCLES(gleam_cpu, 5);
}

void g_php_08()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	stack_write_byte(gleam_cpu.p);
	CYCLES(gleam_cpu, 3);
}

void g_ora_09()
{
	u8 imm = immediate();
#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", imm);
#endif
	gleam_cpu.a |= imm;
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_asl_0a()
{
#ifdef DEBUG
	fprintf(stdout, "A\n");
#endif

	if (gleam_cpu.a & 0x80)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	gleam_cpu.a <<= 1;
	set_flags8(gleam_cpu.a, FL_NEGATIVE | FL_ZERO);
	CYCLES(gleam_cpu, 2);
}

void g_bpl_10()
{
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", nes_read_byte(gleam_cpu.pc));
#endif
	if ((gleam_cpu.p & FL_NEGATIVE) == 0) {
		u16 old = gleam_cpu.pc;
		gleam_cpu.pc = relative();
		if ((old & 0xff00) == (gleam_cpu.pc & 0xff00))
			CYCLES(gleam_cpu, 1);
		else
			CYCLES(gleam_cpu, 2);
	} else
		gleam_cpu.pc++;
	CYCLES(gleam_cpu, 2);
}

void g_asl_16()
{
	u8 imm = immediate();
	u8 val = nes_read_byte((imm & 0xff) + gleam_cpu.x);
#ifdef DEBUG
	fprintf(stdout, "$%02X,X\n", imm);
#endif

	if (val & 0x80)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	val <<= 1;
	set_flags8(val, FL_NEGATIVE | FL_ZERO);
	nes_write_byte((imm & 0xff) + gleam_cpu.x, val);
	CYCLES(gleam_cpu, 6);
}

void g_clc_18()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	SET_FLAG(gleam_cpu, FL_CARRY, 0);
	CYCLES(gleam_cpu, 2);
}

void g_jsr_20()
{				/* TODO: verify */
	stack_write_word(gleam_cpu.pc + 1);	/* stupid processor */
	gleam_cpu.pc = immediate_word();
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", gleam_cpu.pc);
#endif
	CYCLES(gleam_cpu, 6);
}

void g_and_25()
{
	u8 imm = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", imm);
#endif
	gleam_cpu.a &= nes_read_byte(imm & 0xFF);
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 3);
}

void g_plp_28()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.p = stack_read_byte();
	CYCLES(gleam_cpu, 4);
}

void g_and_29()
{
	u8 imm = immediate();
#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", imm);
#endif
	gleam_cpu.a &= imm;
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_rol_2a()
{
	u8 val, c;
#ifdef DEBUG
	fprintf(stdout, "A\n");
#endif
	val = gleam_cpu.a;
	c = (val & 0x80) ? 1 : 0;
	val <<= 1;
	if (gleam_cpu.p & FL_CARRY)
		val |= 0x01;
	if (c)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	gleam_cpu.a = val;
	set_flags8(val, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);

}

void g_bit_2c()
{				/* verify */
	u16 addr = immediate_word();
	u8 val = nes_read_byte(addr);
	u8 tmp;
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", addr);
#endif
	if (val & 0x80)
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 1);
	else
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 0);
	if (val & 0x40)
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 1);
	else
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 0);
	if (val & gleam_cpu.a)
		SET_FLAG(gleam_cpu, FL_ZERO, 0);
	else
		SET_FLAG(gleam_cpu, FL_ZERO, 1);
	CYCLES(gleam_cpu, 4);
}

void g_bmi_30()
{
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", nes_read_byte(gleam_cpu.pc));
#endif
	if ((gleam_cpu.p & FL_NEGATIVE) == 1) {
		u16 old = gleam_cpu.pc;
		gleam_cpu.pc = relative();
		if ((old & 0xff00) == (gleam_cpu.pc & 0xff00))
			CYCLES(gleam_cpu, 1);
		else
			CYCLES(gleam_cpu, 2);
	} else
		gleam_cpu.pc++;
	CYCLES(gleam_cpu, 2);
}

void g_and_35()
{
	u8 imm = immediate();
	u8 val = nes_read_byte((imm & 0xff) + gleam_cpu.x);
#ifdef DEBUG
	fprintf(stdout, "$%02X,X\n", imm);
#endif
	gleam_cpu.a &= val;
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 4);
}

void g_rol_36()
{
	u8 imm = immediate();
	u8 val, c;
#ifdef DEBUG
	fprintf(stdout, "$%02X,X\n", imm);
#endif
	val = nes_read_byte(imm + gleam_cpu.x);
	c = (val & 0x80) >> 7;
	val <<= 1;
	if (gleam_cpu.p & FL_CARRY)
		val |= 0x01;
	if (c)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	nes_write_byte(imm, val);
	set_flags8(val, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 6);
}


void g_sec_38()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	SET_FLAG(gleam_cpu, FL_CARRY, 1);
	CYCLES(gleam_cpu, 2);
}

void g_rts_60()
{				/* TODO: verify */
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.pc = stack_read_word() + 1;
	CYCLES(gleam_cpu, 6);
}

void g_rti_40()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.p = stack_read_byte();
	gleam_cpu.p |= 0x20;	/* r flag, supposed to be 1 always */
	gleam_cpu.pc = stack_read_word();
	CYCLES(gleam_cpu, 6);
}

void g_eor_45()
{
	u8 imm = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", imm);
#endif
	gleam_cpu.a ^= nes_read_byte(imm & 0xFF);
	set_flags8(gleam_cpu.a, FL_NEGATIVE | FL_ZERO);
	CYCLES(gleam_cpu, 3);
}

void g_lsr_46()
{
	u8 imm = immediate();
	u8 val = nes_read_byte(imm & 0xff);
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", imm);
#endif
	if (val & 0x01)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	val >>= 1;
	nes_write_byte(imm & 0xff, val);
	set_flags8(val, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 5);
}

void g_pha_48()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	stack_write_byte(gleam_cpu.a);
	CYCLES(gleam_cpu, 3);
}

void g_eor_49()
{
	u8 imm = immediate();
#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", imm);
#endif
	gleam_cpu.a ^= imm;
	set_flags8(gleam_cpu.a, FL_NEGATIVE | FL_ZERO);
	CYCLES(gleam_cpu, 2);
}

void g_lsr_4a()
{
#ifdef DEBUG
	fprintf(stdout, "A\n");
#endif
	if (gleam_cpu.a & 0x01)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	gleam_cpu.a >>= 1;
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_jmp_4c()
{
	gleam_cpu.pc = immediate_word();
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", gleam_cpu.pc);
#endif
	CYCLES(gleam_cpu, 3);
}

void g_eor_4d()
{
	u16 imm = immediate_word();
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", imm);
#endif
	gleam_cpu.a ^= nes_read_byte(imm);
	set_flags8(gleam_cpu.a, FL_NEGATIVE | FL_ZERO);
	CYCLES(gleam_cpu, 4);
}

void g_pla_68()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.a = stack_read_byte();
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 4);
}


void g_adc_65()
{				/* TODO: verify */
	u8 imm = immediate();
	u8 src = nes_read_byte(imm & 0xff);
	u16 result;

#ifdef DEBUG
	fprintf(stdout, "$%02X\n", imm);
#endif
	result = gleam_cpu.a + ((gleam_cpu.p & FL_CARRY) ? 1 : 0) + src;

	if (!((gleam_cpu.a ^ src) & 0x80)
	    && ((gleam_cpu.a ^ result) & 0x80))
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 1);
	else
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 0);
	if ((gleam_cpu.a & 0x80) && result >= 0x100)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);

	gleam_cpu.a = (result & 0xFF);
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_ror_66()
{
	u8 val, c, imm = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", imm);
#endif
	val = nes_read_byte(imm & 0xFF);
	c = (val & 0x01);
	val >>= 1;
	if (gleam_cpu.p & FL_CARRY)
		val |= 0x80;
	if (c)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	nes_write_byte(imm & 0xFF, val);
	set_flags8(val, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 5);
}

void g_adc_69()
{				/* TODO: verify */
	u8 imm = immediate();
	u16 result;

#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", imm);
#endif

	result = gleam_cpu.a + ((gleam_cpu.p & FL_CARRY) ? 1 : 0) + imm;
	if ((gleam_cpu.a & 0x80) && result >= 0x100)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);

	gleam_cpu.a = (result & 0xFF);
	if (result > 0xFF)
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 1);
	else
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 0);
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_ror_6a()
{
	u8 c;
#ifdef DEBUG
	fprintf(stdout, "A\n");
#endif
	c = (gleam_cpu.a & 0x01);
	gleam_cpu.a >>= 1;
	if (gleam_cpu.p & FL_CARRY)
		gleam_cpu.a |= 0x80;
	if (c)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_adc_6d()
{				/* TODO: verify */
	u16 imm = immediate_word();
	u8 val = nes_read_byte(imm);
	u16 result;

#ifdef DEBUG
	fprintf(stdout, "$%04X\n", imm);
#endif

	result = gleam_cpu.a + ((gleam_cpu.p & FL_CARRY) ? 1 : 0) + val;
	if ((gleam_cpu.a & 0x80) && result >= 0x100)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);

	gleam_cpu.a = (result & 0xFF);
	if (result > 0xFF)
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 1);
	else
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 0);
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_sei_78()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	SET_FLAG(gleam_cpu, FL_INT, 1);
	CYCLES(gleam_cpu, 2);
}

void g_sty_84()
{
	u8 val = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", val);
#endif
	nes_write_byte((val) & 0xFF, gleam_cpu.y);
	CYCLES(gleam_cpu, 3);
}

void g_sta_85()
{
	u8 val = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", val);
#endif
	nes_write_byte((val) & 0xFF, gleam_cpu.a);
	CYCLES(gleam_cpu, 3);
}

void g_stx_86()
{
	u8 val = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", val);
#endif
	nes_write_byte((val) & 0xFF, gleam_cpu.x);
	CYCLES(gleam_cpu, 3);
}

void g_dey_88()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.y--;
	set_flags8(gleam_cpu.y, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_txa_8a()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.a = gleam_cpu.x;
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_sty_8c()
{
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", nes_read_word(gleam_cpu.pc));
#endif
	nes_write_byte(nes_read_word(gleam_cpu.pc), gleam_cpu.y);
	gleam_cpu.pc += 2;
	CYCLES(gleam_cpu, 4);
}

void g_sta_8d()
{
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", nes_read_word(gleam_cpu.pc));
#endif
	nes_write_byte(nes_read_word(gleam_cpu.pc), gleam_cpu.a);
	gleam_cpu.pc += 2;
	CYCLES(gleam_cpu, 4);
}

void g_stx_8e()
{
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", nes_read_word(gleam_cpu.pc));
#endif
	nes_write_byte(nes_read_word(gleam_cpu.pc), gleam_cpu.x);
	gleam_cpu.pc += 2;
	CYCLES(gleam_cpu, 4);
}

void g_bcc_90()
{
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", nes_read_byte(gleam_cpu.pc));
#endif
	if ((gleam_cpu.p & FL_CARRY) == 0) {
		u16 old = gleam_cpu.pc;
		gleam_cpu.pc = relative();
		if ((old & 0xff00) == (gleam_cpu.pc & 0xff00))
			CYCLES(gleam_cpu, 1);
		else
			CYCLES(gleam_cpu, 2);
	} else
		gleam_cpu.pc++;
	CYCLES(gleam_cpu, 2);

}

void g_sta_91()
{
#ifdef DEBUG
	fprintf(stdout, "($%02X),Y\n", nes_read_byte(gleam_cpu.pc));
#endif
	nes_write_byte(post_indexed_indirect(), gleam_cpu.a);
	CYCLES(gleam_cpu, 6);
}

void g_sty_94()
{
	u8 imm = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X,X\n", immediate);
#endif
	nes_write_byte(imm + gleam_cpu.x, gleam_cpu.y);
	CYCLES(gleam_cpu, 4);
}

void g_sta_95()
{
	u8 val = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X,X\n", val);
#endif
	nes_write_byte(val + gleam_cpu.x, gleam_cpu.a);
	CYCLES(gleam_cpu, 4);
}

void g_tya_98()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.a = gleam_cpu.y;
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_sta_99()
{
	u16 val = immediate_word();
#ifdef DEBUG
	fprintf(stdout, "$%04X,Y\n", val);
#endif
	nes_write_byte(val + gleam_cpu.y, gleam_cpu.a);
	CYCLES(gleam_cpu, 5);
}

void g_txs_9a()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.sp = gleam_cpu.x;
	set_flags8(gleam_cpu.sp, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_sta_9d()
{				/* TODO: verify */
	u16 val = immediate_word();
#ifdef DEBUG
	fprintf(stdout, "$%04X,X\n", val);
#endif
	nes_write_byte(val + gleam_cpu.x, gleam_cpu.a);
	CYCLES(gleam_cpu, 5);
}

void g_ldy_a0()
{
	gleam_cpu.y = immediate();
	set_flags8(gleam_cpu.y, FL_ZERO | FL_NEGATIVE);

#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", gleam_cpu.y);
#endif
	CYCLES(gleam_cpu, 2);
}

void g_ldx_a2()
{
	gleam_cpu.x = immediate();
	set_flags8(gleam_cpu.x, FL_ZERO | FL_NEGATIVE);

#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", gleam_cpu.x);
#endif
	CYCLES(gleam_cpu, 2);
}

void g_ldy_a4()
{
	u8 val = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", val);
#endif
	gleam_cpu.y = nes_read_byte(val & 0xff);
	set_flags8(gleam_cpu.y, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 3);
}

void g_lda_a5()
{
	u8 val = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", val);
#endif
	gleam_cpu.a = nes_read_byte(val & 0xff);
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 3);
}

void g_ldx_a6()
{
	gleam_cpu.x = zeropage();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", nes_read_byte(gleam_cpu.pc - 1));
#endif
	set_flags8(gleam_cpu.x, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 3);
}

void g_tay_a8()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.y = gleam_cpu.a;
	set_flags8(gleam_cpu.y, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 4);
}

void g_lda_a9()
{
	gleam_cpu.a = immediate();
#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", gleam_cpu.a);
#endif
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_tax_aa()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.x = gleam_cpu.a;
	set_flags8(gleam_cpu.x, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 4);
}

void g_ldy_ac()
{
	u16 imm = immediate_word();
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", imm);
#endif
	gleam_cpu.y = nes_read_byte(imm);
	set_flags8(gleam_cpu.y, FL_NEGATIVE | FL_ZERO);
	CYCLES(gleam_cpu, 4);
}

void g_lda_ad()
{
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", nes_read_word(gleam_cpu.pc));
#endif
	gleam_cpu.a = absolute();
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 4);
}

void g_ldx_ae()
{
	u16 imm = immediate_word();
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", imm);
#endif
	gleam_cpu.x = nes_read_byte(imm);
	set_flags8(gleam_cpu.x, FL_NEGATIVE | FL_ZERO);
	CYCLES(gleam_cpu, 4);
}

void g_bcs_b0()
{
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", nes_read_byte(gleam_cpu.pc));
#endif
	if ((gleam_cpu.p & FL_CARRY) == 1) {
		u16 old = gleam_cpu.pc;
		gleam_cpu.pc = relative();
		if ((old & 0xff00) == (gleam_cpu.pc & 0xff00))
			CYCLES(gleam_cpu, 1);
		else
			CYCLES(gleam_cpu, 2);
	} else
		gleam_cpu.pc++;
	CYCLES(gleam_cpu, 2);
}

void g_lda_b1()
{
#ifdef DEBUG
	fprintf(stdout, "($%02X),Y\n", nes_read_byte(gleam_cpu.pc));
#endif
	gleam_cpu.a = nes_read_byte(post_indexed_indirect());
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 5);
}

void g_ldy_b4()
{
	u8 val = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X,X\n", val);
#endif
	gleam_cpu.y = nes_read_byte((val & 0xff) + gleam_cpu.x);
	set_flags8(gleam_cpu.y, FL_ZERO | FL_NEGATIVE);

	CYCLES(gleam_cpu, 4);
}

void g_lda_b5()
{
	u8 val = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X,X\n", val);
#endif
	gleam_cpu.a = nes_read_byte((val & 0xff) + gleam_cpu.x);
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);

	if (((u16) (gleam_cpu.x + val)) > 0xFF)
		CYCLES(gleam_cpu, 5);
	else
		CYCLES(gleam_cpu, 4);
}

void g_lda_b9()
{
	u16 val = immediate_word();
#ifdef DEBUG
	fprintf(stdout, "$%04X,Y\n", val);
#endif
	gleam_cpu.a = nes_read_byte(val + gleam_cpu.y);
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);

	if (((u16) (gleam_cpu.y + (val & 0x00FF))) > 0xFF)
		CYCLES(gleam_cpu, 5);
	else
		CYCLES(gleam_cpu, 4);
}

void g_ldy_bc()
{
	u16 val = immediate_word();
#ifdef DEBUG
	fprintf(stdout, "$%04X,X\n", val);
#endif
	gleam_cpu.y = nes_read_byte(val + gleam_cpu.x);
	set_flags8(gleam_cpu.y, FL_ZERO | FL_NEGATIVE);

	if (((val + gleam_cpu.x) & 0xff00) > (gleam_cpu.x & 0xff00))
		CYCLES(gleam_cpu, 1);
	CYCLES(gleam_cpu, 4);
}

void g_ldx_be()
{
	u16 val = immediate_word();
#ifdef DEBUG
	fprintf(stdout, "$%04X,Y\n", val);
#endif
	gleam_cpu.x = nes_read_byte(val + gleam_cpu.y);
	set_flags8(gleam_cpu.x, FL_ZERO | FL_NEGATIVE);

	if (((val + gleam_cpu.y) & 0xff00) > (gleam_cpu.y & 0xff00))
		CYCLES(gleam_cpu, 1);
	CYCLES(gleam_cpu, 4);
}

void g_lda_bd()
{
	u16 addr = immediate_word();
	u16 old = addr;
#ifdef DEBUG
	fprintf(stdout, "$%04X,X\n", addr);
#endif
	addr += gleam_cpu.x;
	/* add 1 if page boundary is crossed */
	if ((old & 0xff00) == ((addr & 0xff00) - 1))
		CYCLES(gleam_cpu, 1);

/* can't believe i wrote this:	nes_write_byte(addr, gleam_cpu.a); */
	gleam_cpu.a = nes_read_byte(addr);
	CYCLES(gleam_cpu, 4);
}

void g_cpy_c0()
{
	u8 bval = immediate();
#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", bval);
#endif
	if (bval > gleam_cpu.y)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	if (bval == gleam_cpu.y)
		SET_FLAG(gleam_cpu, FL_ZERO, 1);
	else
		SET_FLAG(gleam_cpu, FL_ZERO, 0);
	if ((gleam_cpu.y - bval) & 0x80)
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 1);
	else
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 0);
	CYCLES(gleam_cpu, 2);
}

void g_cpy_c4()
{
	u8 bval = zeropage();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", nes_read_byte(gleam_cpu.pc - 1));
#endif
	if (bval > gleam_cpu.y)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	if (bval == gleam_cpu.y)
		SET_FLAG(gleam_cpu, FL_ZERO, 1);
	else
		SET_FLAG(gleam_cpu, FL_ZERO, 0);
	if ((gleam_cpu.y - bval) & 0x80)
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 1);
	else
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 0);
	CYCLES(gleam_cpu, 3);
}

void g_cmp_c5()
{				/* TODO: verify */
	u8 bv = immediate();
	u8 bval = nes_read_byte(bv & 0xff);
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", bv);
#endif
	if (bval > gleam_cpu.a)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	if (bval == gleam_cpu.a)
		SET_FLAG(gleam_cpu, FL_ZERO, 1);
	else
		SET_FLAG(gleam_cpu, FL_ZERO, 0);
	if ((gleam_cpu.a - bval) & 0x80)
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 1);
	else
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 0);
	CYCLES(gleam_cpu, 3);
}

void g_dec_c6()
{
	u8 val2, val = immediate();
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", val);
#endif
	val2 = nes_read_byte((val) & 0xFF) - 1;
	nes_write_byte((val) & 0x0FF, val2);
	set_flags8(val2, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 5);
}

void g_iny_c8()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.y++;
	set_flags8(gleam_cpu.y, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_cmp_c9()
{				/* TODO: verify */
	u8 bval = immediate();
#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", bval);
#endif
	if (bval > gleam_cpu.a)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	if (bval == gleam_cpu.a)
		SET_FLAG(gleam_cpu, FL_ZERO, 1);
	else
		SET_FLAG(gleam_cpu, FL_ZERO, 0);
	if ((gleam_cpu.a - bval) & 0x80)
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 1);
	else
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 0);
	CYCLES(gleam_cpu, 2);
}

void g_dex_ca()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.x--;
	set_flags8(gleam_cpu.x, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_dec_ce()
{
	u16 addr = immediate_word();
	u8 val = nes_read_byte(addr);
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", addr);
#endif
	val--;
	set_flags8(val, FL_ZERO | FL_NEGATIVE);
	nes_write_byte(addr, val);
	CYCLES(gleam_cpu, 6);
}

void g_bne_d0()
{
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", nes_read_byte(gleam_cpu.pc));
	fprintf(stdout, "X = %02X | A = %02X | P = %02X\n", gleam_cpu.x,
		gleam_cpu.a, gleam_cpu.p);
#endif

	if ((gleam_cpu.p & FL_ZERO) == 0) {
		u16 old = gleam_cpu.pc;
		gleam_cpu.pc = relative();
		if ((old & 0xff00) == (gleam_cpu.pc & 0xff00))
			CYCLES(gleam_cpu, 1);
		else
			CYCLES(gleam_cpu, 2);
	} else
		gleam_cpu.pc++;
	CYCLES(gleam_cpu, 2);
}

void g_dec_d6()
{
	u8 imm = immediate();
	u8 val = nes_read_byte((imm & 0xff) + gleam_cpu.x);
#ifdef DEBUG
	fprintf(stdout, "$%02X,X\n", imm);
#endif
	val--;
	set_flags8(val, FL_ZERO | FL_NEGATIVE);
	nes_write_byte((imm & 0xff) + gleam_cpu.x, val);
	CYCLES(gleam_cpu, 6);
}

void g_cld_d8()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	SET_FLAG(gleam_cpu, FL_DEC, 0);
	CYCLES(gleam_cpu, 2);
}

void g_cmp_dd()
{
	u16 bval = immediate_word();
	u8 val = nes_read_byte(bval + gleam_cpu.x);
#ifdef DEBUG
	fprintf(stdout, "$%04X,X\n", bval);
#endif
	if (val > gleam_cpu.a)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	if (val == gleam_cpu.a)
		SET_FLAG(gleam_cpu, FL_ZERO, 1);
	else
		SET_FLAG(gleam_cpu, FL_ZERO, 0);
	if ((gleam_cpu.a - val) & 0x80)
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 1);
	else
		SET_FLAG(gleam_cpu, FL_NEGATIVE, 0);
	if ((bval & 0xff00) != ((bval + gleam_cpu.x) & 0xff00))
		CYCLES(gleam_cpu, 1);
	CYCLES(gleam_cpu, 4);
}

void g_cpx_e0()
{
	u8 imm = immediate();
	u16 res;
#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", imm);
#endif
	res = gleam_cpu.x - imm;
	set_flags8(res, FL_NEGATIVE | FL_ZERO);

	if (imm > gleam_cpu.x)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	CYCLES(gleam_cpu, 2);
}

void g_cpx_e4()
{
	u8 imm = immediate();
	u8 val = nes_read_byte(imm & 0xff);
	u8 res;
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", imm);
#endif
	res = gleam_cpu.x - val;
	set_flags8(res, FL_NEGATIVE | FL_ZERO);

	if (val > gleam_cpu.x)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);
	CYCLES(gleam_cpu, 3);
}

void g_sbc_e5()
{				/* TODO: verify */
	u8 ix = immediate();
	u8 imm = nes_read_byte(ix & 0xff);
	u16 result;

#ifdef DEBUG
	fprintf(stdout, "$%02X\n", ix);
#endif

	result = gleam_cpu.a - imm - ((gleam_cpu.p & FL_CARRY) ? 0 : 1);
	if (((gleam_cpu.a ^ result) & 0x80)
	    && ((gleam_cpu.a ^ imm) & 0x80))
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 1);
	else
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 0);
	if (gleam_cpu.p & FL_DEC) {
		if (((result & 0x0F) -
		     ((gleam_cpu.p & FL_CARRY) ? 0 : 1)) < (imm & 0x0F))
			result -= 6;
		if (result > 0x99)
			result -= 0x60;
	}
	if (result < 0x100)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);

	gleam_cpu.a = result;
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 3);
}

void g_inc_e6()
{
	u8 imm = immediate();
	u8 val;
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", imm);
#endif
	val = nes_read_byte(imm & 0xff) + 1;
	nes_write_byte(imm & 0xff, val);
	set_flags8(val, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 5);
}

void g_inx_e8()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	gleam_cpu.x++;
	set_flags8(gleam_cpu.x, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_sbc_e9()
{				/* TODO: verify */
	u8 imm = immediate();
	u16 result;

#ifdef DEBUG
	fprintf(stdout, "#$%02X\n", imm);
#endif

	result = gleam_cpu.a - imm - ((gleam_cpu.p & FL_CARRY) ? 0 : 1);
	if (((gleam_cpu.a ^ result) & 0x80)
	    && ((gleam_cpu.a ^ imm) & 0x80))
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 1);
	else
		SET_FLAG(gleam_cpu, FL_OVERFLOW, 0);
	if (gleam_cpu.p & FL_DEC) {
		if (((result & 0x0F) -
		     ((gleam_cpu.p & FL_CARRY) ? 0 : 1)) < (imm & 0x0F))
			result -= 6;
		if (result > 0x99)
			result -= 0x60;
	}
	if (result < 0x100)
		SET_FLAG(gleam_cpu, FL_CARRY, 1);
	else
		SET_FLAG(gleam_cpu, FL_CARRY, 0);

	gleam_cpu.a = result;
	set_flags8(gleam_cpu.a, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 2);
}

void g_nop_ea()
{
#ifdef DEBUG
	fprintf(stdout, "\n");
#endif
	CYCLES(gleam_cpu, 2);
}

void g_inc_ee()
{
	u16 addr;
	u8 value;
#ifdef DEBUG
	fprintf(stdout, "$%04X\n", nes_read_word(gleam_cpu.pc));
#endif
	addr = immediate_word();
	value = nes_read_byte(addr) + 1;
	nes_write_byte(addr, value);
	set_flags8(value, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 6);
}

void g_beq_f0()
{
#ifdef DEBUG
	fprintf(stdout, "$%02X\n", nes_read_byte(gleam_cpu.pc));
#endif
	if ((gleam_cpu.p & FL_ZERO) == 1) {
		u16 old = gleam_cpu.pc;
		gleam_cpu.pc = relative();
		if ((old & 0xff00) == (gleam_cpu.pc & 0xff00))
			CYCLES(gleam_cpu, 1);
		else
			CYCLES(gleam_cpu, 2);
	} else
		gleam_cpu.pc++;
	CYCLES(gleam_cpu, 2);
}

void g_inc_fe()
{
	u16 imm = immediate_word();
	u8 val;
#ifdef DEBUG
	fprintf(stdout, "$%04X,X\n", imm);
#endif
	imm += gleam_cpu.x;
	val = nes_read_byte(imm);
	val += 1;
	nes_write_byte(imm, val);
	set_flags8(val, FL_ZERO | FL_NEGATIVE);
	CYCLES(gleam_cpu, 7);
}

void g_unknown()
{
#ifdef TARGET_DC
	dc_vid_border_color(0xFFFF, 0, 0);
#endif
	fprintf(stdout, "UNKNOWN!@#$!@#$\n");
	fprintf(stdout, ":: PC[$%04X] @PC[$%02X] SP[$%04X] P[$%04X]\n",
		gleam_cpu.pc, nes_read_byte(gleam_cpu.pc), gleam_cpu.sp,
		gleam_cpu.p);
	fprintf(stdout, ":: A[$%02X] X[$%02X] Y[$%02X]\n", gleam_cpu.a,
		gleam_cpu.x, gleam_cpu.y);
}

void g_do_nmi()
{
	stack_write_word(gleam_cpu.pc);
	stack_write_byte(gleam_cpu.p);
	gleam_cpu.pc = nes_read_word(NMI_VECTOR);
}
