/* This file is part of the Dreamcast function library.
 * Please see libdream.c for further details.
 *
 * (c)2000 Dan Potter
 */

/* volatile unsigned char *dc_snd_base = (unsigned char *)0x00800000; */
#define dc_snd_base ((volatile unsigned char *)0x00800000)

/* Some convienence macros */
#define SNDREG32A(x) ((volatile unsigned long *)(dc_snd_base + (x)))
#define SNDREG32(x) (*SNDREG32A(x))
#define SNDREG8A(x) (dc_snd_base + (x))
#define SNDREG8(x) (*SNDREG8A(x))
#define CHNREG32A(chn, x) SNDREG32A(0x80*(chn) + (x))
#define CHNREG32(chn, x) (*CHNREG32A(chn, x))
#define CHNREG8A(chn, x) SNDREG8A(0x80*(chn) + (x))
#define CHNREG8(chn, x) (*CHNREG8A(chn, x))

void aica_init() {
	int i, j;
	
	/* Initialize AICA channels */	
	SNDREG32(0x2800) = 0x0000;
	
	for (i=0; i<64; i++) {
		for (j=0; j<0x80; j+=4)
			CHNREG32(i, j) = 0;
		CHNREG32(i,0) = 0x8000;
		CHNREG32(i,20) = 0x1f;
	}

	SNDREG32(0x2800) = 0x000f;
}

/* Sets up a sound channel completely. This is generally good if you want
   a quick and dirty way to play notes. If you want a more comprehensive
   set of routines (more like PC wavetable cards) see below.
   
   ch is the channel to play on (0 - 63)
   smpptr is the pointer to the sound data; if you're running off the
     SH4, then this ought to be (ptr - 0xa0800000); otherwise it's just
     ptr. Basically, it's an offset into sound ram.
   mode is one of the mode constants (16 bit, 8 bit, ADPCM)
   nsamp is the number of samples to play (not number of bytes!)
   freq is the sampling rate of the sound
   vol is the volume, 0 to 0xff (0xff is louder)
   pan is a panning constant -- 0 is left, 128 is center, 255 is right.

   This routine (and the similar ones) owe a lot to Marcus' sound example -- 
   I hadn't gotten quite this far into dissecting the individual regs yet. */
void aica_play(int ch, unsigned long smpptr, int mode, int nsamp, int freq, int vol, int pan) {
	unsigned long freq_lo, freq_base = 5644800;
	int freq_hi = 7;

	/* Convert the incoming volume and pan into hardware values */
	vol = 0xff - (vol & 0xff);
	
	/* For the moment this is going to have to suffice, until we really
	   figure out what these mean. */
	if (pan == 0x80)
		pan = 0;
	else if (pan < 0x80)
		pan = 0x1f;
	else
		pan = 0xf;
	
	/* Set sample format and buffer address */
	CHNREG32(ch, 0) = 0x4200 | (mode<<7) | (smpptr >> 16);
	CHNREG32(ch, 4) = smpptr & 0xffff;
	
	/* Envelope length (we don't use envelopes here) */
	CHNREG32(ch, 8) = (nsamp >> 16) & 0xffff;
	CHNREG32(ch, 12) = nsamp & 0xffff;
	
	/* If we supported volume envelopes (which we don't yet) then
	   this value would set that up. The top 4 bits determine the
	   envelope speed. f is the fastest, 1 is the slowest, and 0
	   seems to be an invalid value and does weird things). The
	   default (below) sets it into normal mode.
	CHNREG32(ch, 16) = 0xf010;
	*/
	
	/* Need to convert frequency to floating point format
	   (freq_hi is exponent, freq_lo is mantissa)
	   Formula is ferq = 44100*2^freq_hi*(1+freq_lo/1024) */
	while (freq < freq_base && freq_hi > -8) {
		freq_base >>= 1;
		--freq_hi;
	}
	freq_lo = (freq<<10) / freq_base;
	
	/* Write resulting values */
	CHNREG32(ch, 24) = (freq_hi << 11) | (freq_lo & 1023);
	
	/* Set volume, pan, and some other things that we don't know what
	   they do =) */
	CHNREG8(ch, 36) = pan;
	CHNREG8(ch, 37) = 0xf;
	CHNREG8(ch, 40) = 0x24;
	CHNREG8(ch, 41) = vol;
	CHNREG32(ch, 16) = 0x1f;
	
	/* Enable playback */
	CHNREG32(ch, 0) |= 0xc000;
}

/* Stop the sound on a given channel */
void aica_stop(int ch) {
	CHNREG32(ch, 0) = (CHNREG32(ch, 0) & ~0x4000) | 0x8000;
}


/* The rest of these routines can change the channel in mid-stride so you
   can do things like vibrato and panning effects. */
   
/* Set channel volume */
void aica_vol(int ch, int vol) {
	vol = 0xff - (vol & 0xff);
	CHNREG8(ch, 41) = vol;
}

/* Set channel pan */
void aica_pan(int ch, int pan) {
	pan = ((pan - 128) >> 3) & 0x1f;
	CHNREG8(ch, 36) = pan;
}

/* Set channel frequency */
void aica_freq(int ch, int freq) {
	unsigned long freq_lo, freq_base = 5644800;
	int freq_hi = 7;

	while (freq < freq_base && freq_hi > -8) {
		freq_base >>= 1;
		freq_hi--;
	}
	freq_lo = (freq<<10) / freq_base;
	CHNREG32(ch, 24) = (freq_hi << 11) | (freq_lo & 1023);
}

