/* This file is part of the Dreamcast function library.
 * Please see libdream.c for further details.
 *
 * (c)2000 Dan Potter
 */

#ifndef __SPU_CONTROL_H
#define __SPU_CONTROL_H

/* This file represents our AICA firmware. Basically it is an SPU
   processing proxy. It waits for commands from the SH4 and executes
   them. For now the only commands we have are to start playing on a
   channel (with given parameters) and to stop playing on a channel.
   
   To execute a command, the SH4 should poke values into control point.
   The control point is a fixed structure with several parameters and
   a "go" flag. The "go" flag will be off by default (to signify that
   the ARM has finished the processing on that command). The SH4 should
   set parameters in the structure and turn on the "go" flag. At this
   point, the ARM will start executing the command. The SH4 should not
   poke any more data here until the "go" flag turns to zero again. It
   is then safe to continue.
   
   In the case of the current structure, the "go" flag is the command slot.
   
   */

typedef struct {
	int 	cmd;		/* Command */
	int	ch;		/* Channel */
	int	pan;		/* Pan value, 0 - 0xff */
	int	vol;		/* Volume value, 0 - 0xff */
	int	freq;		/* Frequency value */
	int	base;		/* Sample base address */
	int	len;		/* Sample length */
	int	mode;		/* Play mode */
} aica_cmd;

#define CMD_NONE	0x0000
#define CMD_PLAY	0x1000
#define CMD_STOP	0x1001
#define CMD_STOPALL	0x1002
#define CMD_SET_PAN	0x2000
#define CMD_SET_VOL	0x2001
#define CMD_SET_FREQ	0x2002


/* Our firmware space ends at 0xb000 (stack top) so we'll put
   the command structures a little after that. */
#define CMD_BASE 0xf000 

/* "Heartbeat" counter -- this counts upwards (and wraps) as long
   as the ARM CPU is running. */
#define HEARTBEAT (*((unsigned long*)CMD_BASE))

/* Command structure -- used for communication with the SH4 */
#define CMD ((volatile aica_cmd*)(CMD_BASE+4))

/* Samples should start at 0x10000 */
#define SMP_BASE 0x10000


#endif	/* __SPU_CONTROL_H */
