/* This file is part of the Dreamcast function library.
 * Please see libdream.c for further details.
 *
 * (c)2000 Dan Potter
 */

#include "dream.h"

/* This include file is generated at compile time; it contains the set of
   firmware that libdream is expecting to see. If you swap out firmware 
   then you can bet on none of these functions working. If you need to
   recompile aica_fw.h, you'll need GCC/Binutils for the ARM7 target. */
#include "aica_fw.h"
#include "aica_fw/spu_control.h"


/* Please note that although the function set here is complete
   for using the Yamaha AICA, you generally will want to load the
   code into the AICA's ARM core to run it. This offloads the processing
   of musical data into another processor and lets you get on with
   whatever you're doing. You don't have to though.
   
   This file is designed so that it will run from either place. You just
   have to call dc_snd_base_arm() if you want it to run from the ARM.

   Note: Currently it doesn't seem to be working from the SH4... I don't
   know why, because my test that manually pokes at the AICA works. ~=) So
   I've moved all the code to its own firmware piece.
*/


/* These constants and variables will determine whether we're running
   off the AICA RISC cpu or the SH4. The default is SH4. */
#define RUNNING_SH4 0
#define RUNNING_ARM 1
volatile unsigned char *dc_snd_base = (unsigned char *)0xa0700000;
static int where = RUNNING_SH4;

/* Some convienence macros */
#define SNDREG32A(x) ((volatile unsigned long *)(dc_snd_base + (x)))
#define SNDREG32(x) (*SNDREG32A(x))
#define SNDREG8A(x) (dc_snd_base + (x))
#define SNDREG8(x) (*SNDREG8A(x))
#define CHNREG32A(chn, x) SNDREG32A(0x80*(chn) + (x))
#define CHNREG32(chn, x) (*CHNREG32A(chn, x))
#define CHNREG8A(chn, x) SNDREG8A(0x80*(chn) + (x))
#define CHNREG8(chn, x) (*CHNREG8A(chn, x))

/* Adapt the SPU addresses to ours */
#define SH_HEARTBEAT (* ((&HEARTBEAT) + 0xa0800000/4) )
#define SH_CMD ((volatile aica_cmd*) (((unsigned char *)CMD) + 0xa0800000))

/******************************************************** RAM management **/

/* Call to run from the ARM */
void dc_snd_base_arm() {
	dc_snd_base = (unsigned char*)0x00800000;
	where = RUNNING_ARM;
}


/* Small utility functions that copy to sound ram using longwords, with
   the delay routine. These only work from the SH4. */
void dc_snd_ram_write_wait() {
	volatile unsigned long *a05f688c = (unsigned long*)0xa05f688c;
	int i;

	for (i=0; i<0x1800; i++) {
		if (a05f688c[0] & 1) break;
	}
}

// length here is longwords, not bytes.
void memcpy4(unsigned long *to, unsigned char *from, int length) {
	unsigned long *froml = (unsigned long*)from;
	int i;
	
	for (i=0; i<length; i++) {
		*to++ = *froml++;
		if (i && !(i % 8)) dc_snd_ram_write_wait();
	}
}

// length here is longwords, not bytes.
void memset4(unsigned long *to, unsigned long what, int length) {
	int i;
	for (i=0; i<length; i++) {
		*to++ = what;
		if (i && !(i % 8)) dc_snd_ram_write_wait();
	}
}

/* Loads a sample into SPU sample ram. This function is generally
   only used from the SH4, so I have left the direct pointer references
   in here. Note that dest is an int, not a pointer. This is because
   it adds the sample ram address for you and you can use the same
   indeces with both load and play. */
void dc_snd_load(void *src, int dest, int len) {
	memcpy4((unsigned long*)(0xa0800000 + dest), src, len/4 + 1);
}

/******************************************************** ARM Control ****/

/* Loads a program into the AICA RISC cpu core and runs it. */
void dc_snd_load_arm(void* src, int size) {
	/* Disable the CPU */
	SNDREG32(0x2c00) |= 1;
	
	/* Clear out sound RAM */
	memset4((void*)0xa0800000, 0, 0x200000/4);
	
	/* Load the program */
	memcpy4((void*)0xa0800000, src, size/4+1);
	
	/* Start the CPU again */
	SNDREG32(0x2c00) &= ~1;
}

/* Halts execution on the ARM processor */
void dc_snd_stop_arm() {
	int i;
	
	/* Stop the ARM processor */
	SNDREG32(0x2c00) |= 1;
	
	/* Make sure we didn't leave any notes running */
	for (i=0; i<64; i++) {
		CHNREG32(i, 0) = (CHNREG32(i, 0) & ~0x4000) | 0x8000;
	}
}

/*********************************************** Firmware Interface ******/

/* Executes a command on the AICA firmware. You should have already set
   everything in the structure except the command code. */
void dc_snd_exec_cmd(int cmd) {
	int spincnt = 0;
	
	SH_CMD->cmd = cmd;
	while ((spincnt++) < 0x100000)
		if (SH_CMD->cmd == CMD_NONE)
			return;
	dc_serial_printf("AICA didn't respond to cmd %d.. is it dead?\r\n", cmd);
}

/* Start a note playing */
void dc_snd_play(int ch, int smpbase, int mode, int nsamp, int freq, int vol, int pan) {
	/* There seem to be some RAM-write timing issues that aren't solved so easily
	   as throwing in a write_wait() call. Humor me for now and leave these in
	   if you want to do any simultaneous or almost-simultaneous starting of
	   sounds on more than one channel. */
	dc_serial_printf("Starting playback on ch %d, smpbase %x, mode %d, nsamp %d, "
		"freq %d, vol %x, pan %x\r\n", ch, smpbase, mode, nsamp, freq, vol, pan);
	SH_CMD->ch = ch;
	SH_CMD->base = smpbase;
	SH_CMD->mode = mode;
	SH_CMD->len = nsamp;
	SH_CMD->freq = freq;
	SH_CMD->vol = vol;
	SH_CMD->pan = pan;
	dc_snd_exec_cmd(CMD_PLAY);
	dc_serial_printf("command is finished\r\n");
}

/* Stop a channel */
void dc_snd_stop(int ch) {
	dc_serial_printf("Stopping playback on ch %d\r\n", ch);
	SH_CMD->ch = ch;
	dc_snd_exec_cmd(CMD_STOP);
}

/* Stop all channels */
void dc_snd_stopall() {
	dc_serial_printf("Stopping all\r\n");
	dc_snd_exec_cmd(CMD_STOPALL);
}

/* Set the volume on a running channel */
void dc_snd_volume(int ch, int vol) {
	dc_serial_printf("Setting volume on %d to %x\r\n", ch, vol);
	SH_CMD->ch = ch;
	SH_CMD->vol = vol;
	dc_snd_exec_cmd(CMD_SET_VOL);
}

/* Set the pan on a running channel */
void dc_snd_pan(int ch, int pan) {
	dc_serial_printf("Setting pan on %d to %x\r\n", ch, pan);
	SH_CMD->ch = ch;
	SH_CMD->pan = pan;
	dc_snd_exec_cmd(CMD_SET_PAN);
}

/* Set the freq on a running channel */
void dc_snd_freq(int ch, int freq) {
	dc_serial_printf("Setting frequency on %d to %x\r\n", ch, freq);
	SH_CMD->ch = ch;
	SH_CMD->freq = freq;
	dc_snd_exec_cmd(CMD_SET_FREQ);
}

/******************************************************** Main init ******/

/* Initialize the AICA sound processor. This loads the firmware code into the
   ARM7 CPU core and starts it executing. It then does a little delay and
   checks to make sure that the CPU is indeed executing =). */
void dc_snd_init() {
	int i, j;

	dc_snd_load_arm(aica_fw, sizeof(aica_fw));

	/* Pause for a moment to let the AICA init */
	dc_sleep(50);
	
	/* Check the heartbeat to make sure it's alive */
	i = SH_HEARTBEAT;
	dc_sleep(50);
	j = SH_HEARTBEAT;
	if (i == j) {
		dc_serial_printf("The AICA seems to be dead.. hmm... (%d=%d)\r\n", i, j);
	}
}




